import time
import numpy as np
import matplotlib.pyplot as plt
from numpy.testing import assert_allclose
from scipy_dae.integrate import solve_dae, consistent_initial_conditions, BDFDAE
from scipy.optimize._numdiff import approx_derivative

mass = 1
gravity = 10


def F(t, vy, vyp):
    y, u, la, mu = vy
    yp, up, lap, mup = vyp

    # prox_N = min(lap, y)
    def fb(a, b):
        return a + b - np.sqrt(a**2 + b**2)
    
    # prox_la = mup * fb(lap, u)
    if y <= 0:
        prox_la = fb(lap, u)
    else:
        prox_la = lap
    prox_mu = fb(mup, y)
    
    return np.array([
                yp - (u + mup),
                mass * up - (-mass * gravity + lap),
                prox_mu,
                prox_la,
        ])


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 3
    t_span = (t0, t1)

    # tolerances
    rtol = atol = 1e2

    # initial conditions
    y0 = np.array([1, 0, 0, 0], dtype=float)
    yp0 = np.array([0, -gravity, 0, 0], dtype=float)

    # F0 = F(t0, y0, yp0)
    # print(f"F0: {F0}")
    # exit()

    ##############
    # dae solution
    ##############
    # method = RadauDAE
    method = BDFDAE
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method, max_step=1e-2)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    tp = t[1:]
    yp = np.diff(y) / np.diff(t)
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # visualization
    fig, ax = plt.subplots(3, 1)

    ax[0].plot(t, y[0], "-ok", label="y")
    ax[0].legend()
    ax[0].grid()

    ax[1].plot(t, y[1], "-ok", label="u")
    ax[1].legend()
    ax[1].grid()

    ax[2].plot(tp, yp[2], "-ok", label="la")
    ax[2].legend()
    ax[2].grid()

    plt.show()



import time
import numpy as np
import matplotlib.pyplot as plt
from scipy_dae.integrate import solve_dae, consistent_initial_conditions
from scipy.optimize._numdiff import approx_derivative


"""Cartesian pendulum, see Hairer1996 Section VII Example 2."""
m = 1
l = 1
g = 10

def F(t, vy, vyp):
    # stabilized index 1
    x, y, u, v, _, _ = vy
    x_dot, y_dot, u_dot, v_dot, la, mu = vyp

    R = np.zeros(6, dtype=np.common_type(vy, vyp))
    R[0] = x_dot - u - 2 * x * mu
    R[1] = y_dot - v - 2 * y * mu
    R[2] = m * u_dot - 2 * x * la
    R[3] = m * v_dot - 2 * y * la + m * g
    R[4] = 2 * x * u + 2 * y * v
    R[5] = x * x + y * y - l * l

    return R

def jac(t, y, yp, f):
    n = len(y)
    z = np.concatenate((y, yp))

    def fun_composite(t, z):
        y, yp = z[:n], z[n:]
        return F(t, y, yp)
    
    J = approx_derivative(lambda z: fun_composite(t, z), 
                            z, method="2-point", f0=f)
    J = J.reshape((n, 2 * n))
    Jy, Jyp = J[:, :n], J[:, n:]
    return Jy, Jyp

def f(t, z):
    y, yp = z[:6], z[6:]
    return np.concatenate((yp, F(t, y, yp)))


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 20
    t_span = (t0, t1)

    # method = "BDF"
    method = "Radau"

    # initial conditions
    y0 = np.array([l, 0, 0, 0, 0, 0], dtype=float)
    yp0 = np.array([0, 0, 0, -g, 0, 0], dtype=float)
    z0 = np.concatenate((y0, yp0))

    yp0 = np.zeros_like(y0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    y0, yp0, fnorm = consistent_initial_conditions(F, jac, t0, y0, yp0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    print(f"fnorm: {fnorm}")

    # solver options
    atol = rtol = 1e-3

    ##############
    # dae solution
    ##############
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    tp = t[1:]
    yp = np.diff(y) / np.diff(t)
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # visualization
    fig, ax = plt.subplots(4, 1)

    ax[0].plot(t, y[0], "-ok", label="x")
    ax[0].plot(t, y[1], "--xk", label="y")
    ax[0].legend()
    ax[0].grid()

    ax[1].plot(t, y[2], "-ok", label="u")
    ax[1].plot(t, y[3], "--xk", label="v")
    ax[1].legend()
    ax[1].grid()

    ax[2].plot(tp, yp[4], "-ok", label="la")
    ax[2].legend()
    ax[2].grid()

    ax[3].plot(tp, yp[5], "--xk", label="mu")
    ax[3].legend()
    ax[3].grid()

    plt.show()



import time
import numpy as np
import matplotlib.pyplot as plt
from numpy.testing import assert_allclose
from scipy_dae.integrate import solve_dae, consistent_initial_conditions, BDFDAE
from scipy.optimize._numdiff import approx_derivative


"""Index3 DAE found in Jay1995 Example 7.1 and 7.2.

References:
-----------
Jay1995: https://doi.org/10.1016/0168-9274(95)00013-K
"""
def F(t, y, yp, nonlinear_multiplier=False):
    y1, y2, z1, z2, _ = y
    y1p, y2p, z1p, z2p, lap = yp

    if nonlinear_multiplier:
        dy3 = -y1 * y2**2 * z2**3 * lap**2
    else:
        dy3 = -y1 * y2**2 * z2**2 * lap

    F = np.zeros(5, dtype=y.dtype)
    F[0] = y1p - (2 * y1 * y2 * z1 * z2)
    F[1] = y2p - (-y1 * y2 * z2**2)
    F[2] = z1p - ((y1 * y2 + z1 * z2) * lap)
    if nonlinear_multiplier:
        F[3] = z2p - (-y1 * y2**2 * z2**3 * lap**2)
    else:
        F[3] = z2p - (-y1 * y2**2 * z2**2 * lap)
    F[4] = y1 * y2**2 - 1.0

    return F


def jac(t, y, yp, f):
    n = len(y)
    z = np.concatenate((y, yp))

    def fun_composite(t, z):
        y, yp = z[:n], z[n:]
        return F(t, y, yp)
    
    J = approx_derivative(lambda z: fun_composite(t, z), 
                            z, method="2-point", f0=f)
    J = J.reshape((n, 2 * n))
    Jy, Jyp = J[:, :n], J[:, n:]
    return Jy, Jyp


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 3
    t_span = (t0, t1)

    # tolerances
    rtol = atol = 1e-5

    # initial conditions
    y0 = np.array([1, 1, 1, 1, 0], dtype=float)
    yp0 = np.array([2, -1, 2, -1, 1], dtype=float)

    # F0 = F(t0, y0, yp0)
    # print(f"F0: {F0}")
    # exit()

    # # TODO: This seems to be wrong here!
    # print(f"y0: {y0}")
    # print(f"yp0: {yp0}")
    # y0, yp0, fnorm = consistent_initial_conditions(F, jac, t0, y0, yp0, fixed_y0=None, fixed_yp0=None)
    # print(f"y0: {y0}")
    # print(f"yp0: {yp0}")
    # print(f"fnorm: {fnorm}")
    # exit()

    ##############
    # dae solution
    ##############
    # method = RadauDAE
    method = BDFDAE
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method, max_order=3)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    tp = t[1:]
    yp = np.diff(y) / np.diff(t)
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # errors
    # assert_allclose(y[0], np.exp(2 * t), rtol=1e-6)
    # assert_allclose(y[1], np.exp(-t), rtol=1e-6)
    # assert_allclose(y[2], np.exp(2 * t), rtol=1e-6)
    # assert_allclose(y[3], np.exp(-t), rtol=1e-6)
    # assert_allclose(y[4], np.exp(t), rtol=1e-6)
    dt = t[1] - t[0]
    error_y1 = np.linalg.norm((y[0] - np.exp(2 * t)) * dt)
    error_y2 = np.linalg.norm((y[1] - np.exp(-t)) * dt)
    error_z1 = np.linalg.norm((y[2] - np.exp(2 * t))  * dt)
    error_z2 = np.linalg.norm((y[3] - np.exp(-t))  * dt)
    error_lap = np.linalg.norm((yp[4] - np.exp(tp))  * dt)
    print(f"error: [{error_y1}, {error_y2}, {error_z1}, {error_z2}, , {error_lap}]")

    # visualization
    fig, ax = plt.subplots()

    ax.plot(t, y[0], "--r", label="y1")
    ax.plot(t, y[1], "--g", label="y2")
    ax.plot(t, y[2], "-.r", label="z1")
    ax.plot(t, y[3], "-.g", label="z2")
    ax.plot(tp, yp[4], "--b", label="u")

    ax.plot(t, np.exp(2 * t), "-r", label="y1/z1 true")
    ax.plot(t, np.exp(-t), "-g", label="y2/z2 true")
    ax.plot(t, np.exp(t), "-b", label="u true")

    ax.grid()
    ax.legend()

    plt.show()



import time
import numpy as np
import matplotlib.pyplot as plt
from numpy.testing import assert_allclose
from scipy_dae.integrate import solve_dae, consistent_initial_conditions
from scipy.optimize._numdiff import approx_derivative


"""Index2 DAE found in Jay1993 Example 7.

References:
-----------
Jay1993: https://link.springer.com/article/10.1007/BF01990349
"""
def F(t, y, yp):
    y1, y2, _ = y
    y1p, y2p, zp = yp

    F = np.zeros(3, dtype=y.dtype)
    # TODO: This multiplier enters nonlinear. Can we cope with that?
    F[0] = y1p - (y1 * y2**2 * zp**2)
    # F[0] = y1p - (y1 * y2**2 * zp)
    F[1] = y2p - (y1**2 * y2**2 - 3 * y2**2 * zp)
    F[2] = y1**2 * y2 - 1.0

    return F


def jac(t, y, yp, f):
    n = len(y)
    z = np.concatenate((y, yp))

    def fun_composite(t, z):
        y, yp = z[:n], z[n:]
        return F(t, y, yp)
    
    J = approx_derivative(lambda z: fun_composite(t, z), 
                            z, method="2-point", f0=f)
    J = J.reshape((n, 2 * n))
    Jy, Jyp = J[:, :n], J[:, n:]
    return Jy, Jyp

jac = None


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 2
    t_span = (t0, t1)

    # tolerances
    rtol = atol = 1e-8

    # initial conditions
    y0 = np.array([1, 1, 0], dtype=float)
    # yp0 = np.array([0, 0, 1], dtype=float) # TODO: Why is this not consistent?
    yp0 = np.array([1, -2, 1], dtype=float)
    # print(F(t0, y0, yp0))
    # exit()

    # # TODO: This seems to be wrong here!
    # print(f"y0: {y0}")
    # print(f"yp0: {yp0}")
    # y0, yp0, fnorm = consistent_initial_conditions(F, jac, t0, y0, yp0, fixed_y0=None, fixed_yp0=None)
    # print(f"y0: {y0}")
    # print(f"yp0: {yp0}")
    # print(f"fnorm: {fnorm}")
    # exit()

    ##############
    # dae solution
    ##############
    method = "Radau"
    # method = "BDF"
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    tp = t[1:]
    yp = np.diff(y) / np.diff(t)
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # errors
    # assert_allclose(y[0], np.exp(t), rtol=1e-8)
    # assert_allclose(y[1], np.exp(-2 * t), rtol=1e-8)
    # assert_allclose(y[2], np.exp(2 * t), rtol=1e-8)
    dt = t[1] - t[0]
    error_y1 = np.linalg.norm((y[0] - np.exp(t)) * dt)
    error_y2 = np.linalg.norm((y[1] - np.exp(-2 * t)) * dt)
    # Note: We have an expected order reduction of h here since we control the 
    #       error of y[2] instead of yp[2].
    # error_y3 = np.linalg.norm((yp[2] - np.exp(2 * tp))  * dt)
    error_y3 = np.linalg.norm((yp[2] - np.exp(2 * tp))  * dt**2)
    print(f"error: [{error_y1}, {error_y2}, {error_y3}]")

    # visualization
    fig, ax = plt.subplots()

    ax.plot(t, y[0], "--r", label="y1")
    ax.plot(t, y[1], "--g", label="y2")
    ax.plot(tp, yp[2], "--b", label="z")

    ax.plot(t, np.exp(t), "-r", label="y1 true")
    ax.plot(t, np.exp(-2 * t), "-g", label="y2 true")
    ax.plot(t, np.exp(2 * t), "-b", label="z true")

    ax.grid()
    ax.legend()

    plt.show()



from itertools import product
import numpy as np
from numpy.testing import (assert_, assert_allclose,
                           assert_equal, assert_no_warnings, suppress_warnings)
import pytest
from scipy.sparse import coo_matrix, csc_matrix, diags, identity
from scipy_dae.integrate import solve_dae

from scipy.integrate._ivp.tests.test_ivp import fun_linear, jac_linear, sol_linear
from scipy.integrate._ivp.tests.test_ivp import fun_rational, fun_rational_vectorized, jac_rational, jac_rational_sparse, sol_rational
from scipy.integrate._ivp.tests.test_ivp import fun_complex, jac_complex, jac_complex_sparse, sol_complex


def compute_error(y, y_true, rtol, atol):
    e = (y - y_true) / (atol + rtol * np.abs(y_true))
    return np.linalg.norm(e, axis=0) / np.sqrt(e.shape[0])


def F_linear(t, y, yp):
    return yp - fun_linear(t, y)


J_linear = (
    -jac_linear(),
    np.eye(2),
)


J_linear_sparse = (
    -csc_matrix(jac_linear()),
    identity(2, format="csc"),
)


def F_rational(t, y, yp):
    return yp - fun_rational(t, y)


def F_rational_vectorized(t, y, yp):
    return yp - fun_rational_vectorized(t, y)


def J_rational(t, y, yp):
    Jy = -jac_rational(t, y)
    Jyp = np.eye(2)
    return Jy, Jyp


def J_rational_sparse(t, y, yp):
    Jy = -jac_rational_sparse(t, y)
    Jyp = identity(2)
    return Jy, Jyp


def F_complex(t, y, yp):
    return yp - fun_complex(t, y)


def J_complex(t, y, yp):
    Jy = -jac_complex(t, y)
    Jyp = np.eye(1)
    return Jy, Jyp


def J_complex_sparse(t, y, yp):
    Jy = -jac_complex_sparse(t, y)
    Jyp = identity(1, dtype=np.common_type(y, yp))
    return Jy, Jyp


parameters_linear = product(
    ["BDF", "Radau"], # method
    [None, J_linear, J_linear_sparse] # jac
)
@pytest.mark.parametrize("method, jac", parameters_linear)
def test_integration_const_jac(method, jac):
    rtol = 1e-3
    atol = 1e-6
    y0 = [0, 2]
    yp0 = fun_linear(0, y0)
    t_span = [0, 2]

    res = solve_dae(F_linear, t_span, y0, yp0, rtol=rtol, atol=atol, 
                    method=method, dense_output=True, jac=jac)
    
    assert_equal(res.t[0], t_span[0])
    assert_(res.t_events is None)
    assert_(res.y_events is None)
    assert_(res.success)
    assert_equal(res.status, 0)

    y_true = sol_linear(res.t)
    e = compute_error(res.y, y_true, rtol, atol)
    assert_(np.all(e < 5))


parameters_complex = product(
    ["BDF"], # method
    [None, J_complex, J_complex_sparse] # jac
)
@pytest.mark.parametrize("method, jac", parameters_complex)
def test_integration_complex(method, jac):
    rtol = 1e-3
    atol = 1e-6
    y0 = np.array([0.5 + 1j])
    yp0 = fun_complex(0, y0)
    t_span = [0, 1]
    tc = np.linspace(t_span[0], t_span[1])

    res = solve_dae(F_complex, t_span, y0, yp0, rtol=rtol, atol=atol, 
                    method=method, dense_output=True, jac=jac)
    
    assert_equal(res.t[0], t_span[0])
    assert_(res.t_events is None)
    assert_(res.y_events is None)
    assert_(res.success)
    assert_equal(res.status, 0)

    # assert res.nfev < 25
    assert res.nfev < 29

    if method == 'BDF':
        assert_equal(res.njev, 1)
        assert res.nlu < 6
    else:
        assert res.njev == 0
        assert res.nlu == 0

    y_true = sol_complex(res.t)
    e = compute_error(res.y, y_true, rtol, atol)
    assert np.all(e < 5)

    yc_true = sol_complex(tc)
    yc = res.sol(tc)
    e = compute_error(yc, yc_true, rtol, atol)

    assert np.all(e < 5)


parameters_rational = product(
    [False], # vectorized
    ["BDF", "Radau"], # method
    [[5, 9], [5, 1]], # t_span
    [None, J_rational, J_rational_sparse] # jac
)
@pytest.mark.parametrize("vectorized, method, t_span, jac", parameters_rational)
def test_integration_rational(vectorized, method, t_span, jac):
    rtol = 1e-3
    atol = 1e-6
    y0 = [1/3, 2/9]
    yp0 = fun_rational(5, y0)

    if vectorized:
        fun = F_rational_vectorized
    else:
        fun = F_rational

    res = solve_dae(fun, t_span, y0, yp0, rtol=rtol, atol=atol, 
                    method=method, dense_output=True, jac=jac, 
                    vectorized=vectorized)
    
    assert_equal(res.t[0], t_span[0])
    assert_(res.t_events is None)
    assert_(res.y_events is None)
    assert_(res.success)
    assert_equal(res.status, 0)

    if method == "BDF":
        assert_(0 < res.njev < 3)
    else: # Radau
        assert_(0 < res.njev < 4)
    assert_(0 < res.nlu < 10)

    y_true = sol_rational(res.t)
    e = compute_error(res.y, y_true, rtol, atol)
    assert_(np.all(e < 6))

    tc = np.linspace(*t_span)
    yc_true = sol_rational(tc)
    yc = res.sol(tc)

    e = compute_error(yc, yc_true, rtol, atol)
    assert_(np.all(e < 6))

    tc = (t_span[0] + t_span[-1]) / 2
    yc_true = sol_rational(tc)
    yc = res.sol(tc)

    e = compute_error(yc, yc_true, rtol, atol)
    assert_(np.all(e < 5))

    assert_allclose(res.sol(res.t), res.y, rtol=1e-15, atol=1e-15)


parameters_stiff = ["BDF", "Radau"]
@pytest.mark.slow
@pytest.mark.parametrize("method", parameters_stiff)
def test_integration_stiff(method):
    def fun_robertson(t, state):
        x, y, z = state
        return [
            -0.04 * x + 1e4 * y * z,
            0.04 * x - 1e4 * y * z - 3e7 * y * y,
            3e7 * y * y,
        ]
    
    def F_robertson(t, state, statep):
        return statep - fun_robertson(t, state)
    
    rtol = 1e-6
    atol = 1e-6
    y0 = [1e4, 0, 0]
    yp0 = fun_robertson(0, y0)
    tspan = [0, 1e8]

    if method == "BDF":
        for NDF_strategy, max_order in product(
            ["stability", "efficiency", None], # NDF_strategy
            [1, 2, 3, 4, 5, 6], # max_order
        ):
            with suppress_warnings() as sup:
                sup.filter(UserWarning,
                        "Choosing `max_order = 6` is not recomended due to its "
                        "poor stability properties.")
                res = solve_dae(F_robertson, tspan, y0, yp0, rtol=rtol,
                                atol=atol, method=method, max_order=max_order,
                                NDF_strategy=NDF_strategy)
    else: # Radau
        res = solve_dae(F_robertson, tspan, y0, yp0, rtol=rtol,
                        atol=atol, method=method)

    # If the stiff mode is not activated correctly, these numbers will be much 
    # bigger (see max_order=1 case)
    if method == "BDF" and max_order == 1:
        assert res.nfev < 21000
    else:
        assert res.nfev < 5000
    assert res.njev < 200
    
if __name__ == "__main__":
    for params in parameters_linear:
        test_integration_const_jac(*params)

    for params in parameters_complex:
        test_integration_complex(*params)

    for params in parameters_rational:
        test_integration_rational(*params)

    for params in parameters_stiff:
        test_integration_stiff(*params)



import numpy as np
from warnings import warn
from scipy.integrate._ivp.common import norm, EPS, warn_extraneous
from scipy.integrate._ivp.base import DenseOutput
from .dae import DaeSolver


NEWTON_MAXITER = 4
MIN_FACTOR = 0.2
MAX_FACTOR = 10


def compute_R(order, factor):
    """Compute the matrix for changing the differences array."""
    I = np.arange(1, order + 1)[:, None]
    J = np.arange(1, order + 1)
    M = np.zeros((order + 1, order + 1))
    M[1:, 1:] = (I - 1 - factor * J) / I
    M[0] = 1
    return np.cumprod(M, axis=0)


def change_D(D, order, factor):
    """Change differences array in-place when step size is changed."""
    R = compute_R(order, factor)
    U = compute_R(order, 1)
    RU = R.dot(U)
    D[:order + 1] = np.dot(RU.T, D[:order + 1])


def solve_bdf_system(fun, t_new, y_predict, c, psi, LU, solve_lu, scale, tol):
    """Solve the algebraic system resulting from BDF method."""
    d = np.zeros_like(y_predict)
    y = y_predict.copy()
    dy_norm_old = None
    converged = False
    for k in range(NEWTON_MAXITER):
        yp = c * d + psi
        f = fun(t_new, y, yp)
        if not np.all(np.isfinite(f)):
            break

        dy = solve_lu(LU, -f)
        dy_norm = norm(dy / scale)

        if dy_norm_old is None:
            rate = None
        else:
            rate = dy_norm / dy_norm_old

        if (rate is not None and (rate >= 1 or rate ** (NEWTON_MAXITER - k) / (1 - rate) * dy_norm > tol)):
            break

        y += dy
        d += dy

        if (dy_norm == 0 or rate is not None and rate / (1 - rate) * dy_norm < tol):
            converged = True
            break

        dy_norm_old = dy_norm

    return converged, k + 1, y, yp, d


# TODO:
# - adapt documentation
# - add consistent initial conditions somehow
# - add dense output for yp
class BDFDAE(DaeSolver):
    """Implicit method based on backward-differentiation formulas.

    This is a variable order method with the order varying automatically from
    1 to 5. The general framework of the BDF algorithm is described in [1]_.
    This class implements a quasi-constant step size as explained in [2]_.
    The error estimation strategy for the constant-step BDF is derived in [3]_.

    Different numerical differentiation formulas (NDF) are implemented. The 
    choice of [5]_ enhances the stability, while [2]_ improves the accuracy 
    of the method. Standard BDF methods are also implemented, although the 
    first and second order formula use the accuracy enhancement of [2]_ and 
    [5]_ since both methods are L-stable.

    Can be applied in the complex domain.

    Parameters
    ----------
    fun : callable
        Function defining the DAE system: ``f(t, y, yp) = 0``. The calling 
        signature is ``fun(t, y, yp)``, where ``t`` is a scalar and 
        ``y, yp`` are ndarrays with 
        ``len(y) = len(yp) = len(y0) = len(yp0)``. ``fun`` must return 
        an array of the same shape as ``y, yp``. See `vectorized` for more
        information.
    t0 : float
        Initial time.
    y0 : array_like, shape (n,)
        Initial state.
    yp0 : array_like, shape (n,)
        Initial derivative.
    t_bound : float
        Boundary time - the integration won't continue beyond it. It also
        determines the direction of the integration.
    first_step : float or None, optional
        Initial step size. Default is ``None`` which means that the algorithm
        should choose.
    max_step : float, optional
        Maximum allowed step size. Default is np.inf, i.e., the step size is not
        bounded and determined solely by the solver.
    rtol, atol : float and array_like, optional
        Relative and absolute tolerances. The solver keeps the local error
        estimates less than ``atol + rtol * abs(y)``. Here `rtol` controls a
        relative accuracy (number of correct digits), while `atol` controls
        absolute accuracy (number of correct decimal places). To achieve the
        desired `rtol`, set `atol` to be smaller than the smallest value that
        can be expected from ``rtol * abs(y)`` so that `rtol` dominates the
        allowable error. If `atol` is larger than ``rtol * abs(y)`` the
        number of correct digits is not guaranteed. Conversely, to achieve the
        desired `atol` set `rtol` such that ``rtol * abs(y)`` is always smaller
        than `atol`. If components of y have different scales, it might be
        beneficial to set different `atol` values for different components by
        passing array_like with shape (n,) for `atol`. Default values are
        1e-3 for `rtol` and 1e-6 for `atol`.
    jac : (array_like, array_like), (sparse_matrix, sparse_matrix), callable or None, optional
        Jacobian matrices of the right-hand side of the system with respect
        to y and y'. The Jacobian matrices have shape (n, n) and their 
        elements (i, j) are equal to ``d f_i / d y_j`` and 
        ``d f_i / d y_j'``, respectively.  There are three ways to define 
        the Jacobian:

            * If (array_like, array_like) or (sparse_matrix, sparse_matrix) 
              the Jacobian matrices are assumed to be constant.
            # TODO: Add constant J_y'!
            * If callable, the Jacobians are assumed to depend on t, y and y'; 
              it will be called as ``jac(t, y, y')``, as necessary. Additional 
              arguments have to be passed if ``args`` is used (see 
              documentation of ``args`` argument). The return values might be 
              a tuple of sparse matrices.
            * If None (default), the Jacobians will be approximated by finite 
              differences.

        It is generally recommended to provide the Jacobians rather than
        relying on a finite-difference approximation.
    # TODO: Adapt and test this.
    jac_sparsity : {None, array_like, sparse matrix}, optional
        Defines a sparsity structure of the Jacobian matrix for a
        finite-difference approximation. Its shape must be (n, n). This argument
        is ignored if `jac` is not `None`. If the Jacobian has only few non-zero
        elements in *each* row, providing the sparsity structure will greatly
        speed up the computations [4]_. A zero entry means that a corresponding
        element in the Jacobian is always zero. If None (default), the Jacobian
        is assumed to be dense.
    # TODO: Adapt and test this
    vectorized : bool, optional
        Whether `fun` can be called in a vectorized fashion. Default is False.

        If ``vectorized`` is False, `fun` will always be called with ``y`` of
        shape ``(n,)``, where ``n = len(y0)``.

        If ``vectorized`` is True, `fun` may be called with ``y`` of shape
        ``(n, k)``, where ``k`` is an integer. In this case, `fun` must behave
        such that ``fun(t, y)[:, i] == fun(t, y[:, i])`` (i.e. each column of
        the returned array is the time derivative of the state corresponding
        with a column of ``y``).

        Setting ``vectorized=True`` allows for faster finite difference
        approximation of the Jacobian by this method, but may result in slower
        execution overall in some circumstances (e.g. small ``len(y0)``).

    Attributes
    ----------
    n : int
        Number of equations.
    status : string
        Current status of the solver: 'running', 'finished' or 'failed'.
    t_bound : float
        Boundary time.
    direction : float
        Integration direction: +1 or -1.
    t : float
        Current time.
    y : ndarray
        Current state.
    t_old : float
        Previous time. None if no steps were made yet.
    step_size : float
        Size of the last successful step. None if no steps were made yet.
    nfev : int
        Number of evaluations of the right-hand side.
    njev : int
        Number of evaluations of the Jacobian.
    nlu : int
        Number of LU decompositions.

    References
    ----------
    .. [1] G. D. Byrne, A. C. Hindmarsh, "A Polyalgorithm for the Numerical
           Solution of Ordinary Differential Equations", ACM Transactions on
           Mathematical Software, Vol. 1, No. 1, pp. 71-96, March 1975.
    .. [2] L. F. Shampine, M. W. Reichelt, "THE MATLAB ODE SUITE", SIAM J. SCI.
           COMPUTE., Vol. 18, No. 1, pp. 1-22, January 1997.
    .. [3] E. Hairer, G. Wanner, "Solving Ordinary Differential Equations I:
           Nonstiff Problems", Sec. III.2.
    .. [4] A. Curtis, M. J. D. Powell, and J. Reid, "On the estimation of
           sparse Jacobian matrices", Journal of the Institute of Mathematics
           and its Applications, 13, pp. 117-120, 1974.
    .. [5] R. W. Klopfenstein, "Numerical differentiation formulas for stiff 
           systems of ordinary differential equations", RCA Review, 32, 
           pp. 447-462, September 1971.
    """
    def __init__(self, fun, t0, y0, yp0, t_bound, max_step=np.inf,
                 rtol=1e-3, atol=1e-6, jac=None, jac_sparsity=None,
                 vectorized=False, first_step=None, max_order=5,
                 NDF_strategy="stability", **extraneous):
        warn_extraneous(extraneous)
        super().__init__(fun, t0, y0, yp0, t_bound, rtol, atol, 
                         first_step, max_step, vectorized, jac, 
                         jac_sparsity, support_complex=True)
        
        self.newton_tol = max(10 * EPS / rtol, min(0.03, rtol ** 0.5))

        assert 1 <= max_order <= 6, "Ensure that 1 <= max_order <= 6."
        if max_order == 6:
            warn("Choosing `max_order = 6` is not recomended due to its poor stability properties.",
                 stacklevel=3)
        self.max_order = max_order

        if NDF_strategy == "stability":
            # Increase A(alpha) stability without decreasing efficiency  
            # too much. This uses the coefficients of [5] but also enhances 
            # the first order coefficient as proposed in [2].
            kappa = np.array([0, -37 / 200, -1/9, 0.0834, 0.0665, 0.0551, 0.0464])[:max_order + 1]
        elif NDF_strategy == "efficiency":
            # Increase efficiency without decreasing A(alpha) stability 
            # too much, see [2].
            kappa = np.array([0, -37 / 200, -1/9, -0.0823, -0.0415, 0, 0])[:max_order + 1]
        else:
            # BDF case with improved efficiency for first and second order 
            # method as proposed in [2] and [5].
            kappa = np.array([0, -37 / 200, -1/9, 0, 0, 0, 0])[:max_order + 1]

        self.gamma = np.hstack((0, np.cumsum(1 / np.arange(1, max_order + 1))))
        self.alpha = (1 - kappa) * self.gamma
        self.error_const = kappa * self.gamma + 1 / np.arange(1, max_order + 2)

        D = np.empty((max_order + 3, self.n), dtype=self.y.dtype)
        D[0] = self.y
        D[1] = self.yp * self.h_abs * self.direction
        self.D = D

        self.order = 1
        self.n_equal_steps = 0
        self.LU = None

    def _step_impl(self):
        t = self.t
        D = self.D
        y = self.y
        yp = self.yp
        f = self.f

        max_step = self.max_step
        min_step = 10 * np.abs(np.nextafter(t, self.direction * np.inf) - t)
        if self.h_abs > max_step:
            h_abs = max_step
            change_D(D, self.order, max_step / self.h_abs)
            self.n_equal_steps = 0
        elif self.h_abs < min_step:
            h_abs = min_step
            change_D(D, self.order, min_step / self.h_abs)
            self.n_equal_steps = 0
        else:
            h_abs = self.h_abs

        # self.hs.append(h_abs)
        # self.orders.append(self.order)
        # print(f"- t: {t:.3e}; h: {h_abs:.3e}; order: {self.order}")

        atol = self.atol
        rtol = self.rtol
        order = self.order

        alpha = self.alpha
        gamma = self.gamma
        error_const = self.error_const

        Jy = self.Jy
        Jyp = self.Jyp
        LU = self.LU
        current_jac = self.jac is None

        step_accepted = False
        while not step_accepted:
            if h_abs < min_step:
                return False, self.TOO_SMALL_STEP

            h = h_abs * self.direction
            t_new = t + h

            if self.direction * (t_new - self.t_bound) > 0:
                t_new = self.t_bound
                change_D(D, order, np.abs(t_new - t) / h_abs)
                self.n_equal_steps = 0
                LU = None

            h = t_new - t
            h_abs = np.abs(h)

            y_predict = np.sum(D[:order + 1], axis=0)

            scale = atol + rtol * np.abs(y_predict)
            psi = np.dot(D[1: order + 1].T, gamma[1: order + 1]) / h

            converged = False
            c = alpha[order] / h
            while not converged:
                if LU is None:
                    LU = self.lu(Jy + c * Jyp)

                converged, n_iter, y_new, yp_new, d = solve_bdf_system(
                    self.fun, t_new, y_predict, c, psi, 
                    LU, self.solve_lu, scale, self.newton_tol)

                if not converged:
                    if current_jac:
                        break
                    Jy, Jyp = self.jac(t, y, yp, f)
                    LU = None
                    current_jac = True

            if not converged:
                factor = 0.5
                h_abs *= factor
                change_D(D, order, factor)
                self.n_equal_steps = 0
                LU = None
                continue

            safety = 0.9 * (2 * NEWTON_MAXITER + 1) / (2 * NEWTON_MAXITER + n_iter)

            error = error_const[order] * d
            scale = atol + rtol * np.abs(y_new)
            error_norm = norm(error / scale)

            if error_norm > 1:
                factor = max(MIN_FACTOR,
                             safety * error_norm ** (-1 / (order + 1)))
                h_abs *= factor
                change_D(D, order, factor)
                self.n_equal_steps = 0
                # As we didn't have problems with convergence, we don't
                # reset LU here.
            else:
                step_accepted = True

        self.n_equal_steps += 1

        self.t = t_new
        self.y = y_new
        self.yp = yp_new

        self.h_abs = h_abs
        self.Jy = Jy
        self.Jyp = Jyp
        self.LU = LU

        f_new = self.fun(t_new, y_new, yp_new)
        self.f = f_new

        # Update differences. The principal relation here is
        # D^{j + 1} y_n = D^{j} y_n - D^{j} y_{n - 1}. Keep in mind that D
        # contained difference for previous interpolating polynomial and
        # d = D^{k + 1} y_n. Thus this elegant code follows.
        D[order + 2] = d - D[order + 1]
        D[order + 1] = d
        for i in reversed(range(order + 1)):
            D[i] += D[i + 1]

        if self.n_equal_steps < order + 1:
            return True, None

        if order > 1:
            error_m = error_const[order - 1] * D[order]
            error_m_norm = norm(error_m / scale)
        else:
            error_m_norm = np.inf

        if order < self.max_order:
            error_p = error_const[order + 1] * D[order + 2]
            error_p_norm = norm(error_p / scale)
        else:
            error_p_norm = np.inf

        error_norms = np.array([error_m_norm, error_norm, error_p_norm])
        with np.errstate(divide='ignore'):
            factors = error_norms ** (-1 / np.arange(order, order + 3))

        # choose order with largest factor
        delta_order = np.argmax(factors) - 1

        # choose with smallest error
        # TODO: This choice is advertaised in Shampine2002 but experminets 
        # indicate it is not worth it
        # delta_order = np.argmin(error_norms) - 1

        order += delta_order
        self.order = order

        factor = min(MAX_FACTOR, safety * np.max(factors))
        self.h_abs *= factor
        change_D(D, order, factor)
        self.n_equal_steps = 0
        self.LU = None

        return True, None

    def _dense_output_impl(self):
        return BdfDenseOutput(self.t_old, self.t, self.h_abs * self.direction,
                              self.order, self.D[:self.order + 1].copy())


class BdfDenseOutput(DenseOutput):
    def __init__(self, t_old, t, h, order, D):
        super().__init__(t_old, t)
        self.order = order
        self.t_shift = self.t - h * np.arange(self.order)
        self.denom = h * (1 + np.arange(self.order))
        self.D = D

    def _call_impl(self, t):
        if t.ndim == 0:
            x = (t - self.t_shift) / self.denom
            p = np.cumprod(x)
        else:
            x = (t - self.t_shift[:, None]) / self.denom[:, None]
            p = np.cumprod(x, axis=0)

        y = np.dot(self.D[1:].T, p)
        if y.ndim == 1:
            y += self.D[0]
        else:
            y += self.D[0, :, None]

        return y



import time
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import solve_ivp
from scipy_dae.integrate import solve_dae, consistent_initial_conditions
from scipy.optimize._numdiff import approx_derivative


"""Robertson problem of semi-stable chemical reaction, see mathworks and Shampine2005.

References:
-----------
mathworks: https://de.mathworks.com/help/matlab/math/solve-differential-algebraic-equations-daes.html#bu75a7z-5 \\
Shampine2005: https://doi.org/10.1016/j.amc.2004.12.011
"""

def f(t, y):
    y1, y2, y3 = y

    yp = np.zeros(3, dtype=y.dtype)
    yp[0] = -0.04 * y1 + 1e4 * y2 * y3
    yp[1] = 0.04 * y1 - 1e4 * y2 * y3 - 3e7 * y2**2
    yp[2] = 3e7 * y2**2

    return yp

def F(t, y, yp):
    y1, y2, y3 = y
    y1p, y2p, y3p = yp

    F = np.zeros(3, dtype=y.dtype)
    F[0] = y1p - (-0.04 * y1 + 1e4 * y2 * y3)
    F[1] = y2p - (0.04 * y1 - 1e4 * y2 * y3 - 3e7 * y2**2)
    F[2] = y1 + y2 + y3 - 1

    return F

def jac(t, y, yp, f):
    n = len(y)
    z = np.concatenate((y, yp))

    def fun_composite(t, z):
        y, yp = z[:n], z[n:]
        return F(t, y, yp)
    
    J = approx_derivative(lambda z: fun_composite(t, z), 
                            z, method="2-point", f0=f)
    J = J.reshape((n, 2 * n))
    Jy, Jyp = J[:, :n], J[:, n:]
    return Jy, Jyp


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 1e7
    t_span = (t0, t1)

    # method = "BDF"
    method = "Radau"

    # initial conditions
    y0 = np.array([1, 0, 0], dtype=float)
    yp0 = f(t0, y0)

    yp0 = np.zeros_like(y0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    y0, yp0, fnorm = consistent_initial_conditions(F, jac, t0, y0, yp0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    print(f"fnorm: {fnorm}")

    # solver options
    atol = rtol = 1e-6

    ####################
    # reference solution
    ####################
    start = time.time()
    sol = solve_ivp(f, t_span, y0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t_scipy = sol.t
    y_scipy = sol.y
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    ##############
    # dae solution
    ##############
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # visualization
    fig, ax = plt.subplots()

    ax.plot(t, y[0], "-ok", label="y1 DAE" + f" ({method})", mfc="none")
    ax.plot(t, y[1] * 1e4, "-ob", label="y2 DAE" + f" ({method})", mfc="none")
    ax.plot(t, y[2], "-og", label="y3 DAE" + f" ({method})", mfc="none")
    ax.plot(t_scipy, y_scipy[0], "xr", label="y1 scipy Radau", markersize=7)
    ax.plot(t_scipy, y_scipy[1] * 1e4, "xy", label="y2 scipy Radau", markersize=7)
    ax.plot(t_scipy, y_scipy[2], "xm", label="y3 scipy Radau", markersize=7)
    ax.set_xscale("log")
    ax.legend()
    ax.grid()

    plt.show()



######################################
# derived from
# from scipy.integrate._ivp import ivp
######################################

import inspect
import numpy as np
from scipy.integrate._ivp.ivp import OdeResult, prepare_events, handle_events, find_active_events
from scipy.integrate._ivp.common import OdeSolution
from .base import DaeSolver
from .bdf import BDFDAE
from .radau import RadauDAE


METHODS = {
    "BDF": BDFDAE,
    "Radau": RadauDAE,
}


MESSAGES = {0: "The solver successfully reached the end of the integration interval.",
            1: "A termination event occurred."}


# TODO:
# - expect consistent initial conditions and add a helper function that computes them as done by matlab?
# - add events depending on y'(t)?
# - return y' and interpolate y' somehow?
def solve_dae(fun, t_span, y0, y_dot0, method="BDF", t_eval=None, 
              dense_output=False, events=None, vectorized=False, 
              args=None, **options):
    """Solve an initial value problem for a system of differential algebraic 
    equations (DAE's).

    This function numerically integrates a system of implicit ordinary 
    differential equations given an initial value::

        f(t, y, y') = 0
        y(t0) = y0
        y'(t0) = y_dot0

    Here t is a 1-D independent variable (time), y(t) is an N-D vector-valued 
    function (state), y'(t) is an N-D vector-valued function (state 
    derivative) and an N-D vector-valued function f(t, y, y') determines the 
    implicit differential algebraic equations.
    The goal is to find y(t) and y'(t) approximately satisfying the differential
    algebraic equations, given initial values y(t0)=y0 and y'(t0)=y_dot0.

    Some of the solvers support integration in the complex domain, but note
    that the function f must be complex-differentiable (satisfy Cauchy-Riemann 
    equations [11]_).
    To solve a problem in the complex domain, pass y0 or y_dot0 with a complex 
    data type. Another option always available is to rewrite your problem for 
    real and imaginary parts separately.

    Parameters
    ----------
    fun : callable
        Function defining the DAE system: ``f(t, y, y_dot) = 0``. The calling 
        signature is ``fun(t, y, y_dot)``, where ``t`` is a scalar and 
        ``y, y_dot`` are ndarrays with 
        ``len(y) = len(y_dot) = len(y0) = len(y_dot0)``. Additional 
        arguments need to be passed if ``args`` is used (see documentation of
        ``args`` argument). ``fun`` must return an array of the same shape as
        ``y`` and ``y_dot``. See `vectorized` for more information.
    t_span : 2-member sequence
        Interval of integration (t0, t_bound). The solver starts with t=t0 and
        integrates until it reaches t=t_bound. Both t0 and t_bound must be 
        floats or values interpretable by the float conversion function.
    y0 : array_like, shape (n,)
        Initial state. For problems in the complex domain, pass `y0` with a
        complex data type (even if the initial value is purely real).
    y_dot0 : array_like, shape (n,)
        Initial derivative. For problems in the complex domain, pass `y_dot0` 
        with a complex data type (even if the initial value is purely real).
    method : string or `DaeSolver`, optional
        Integration method to use:

            * 'Radau': Implicit Runge-Kutta method of the Radau IIA family of
              order 5 [4]_. The error is controlled with a third-order accurate
              embedded formula. A cubic polynomial which satisfies the
              collocation conditions is used for the dense output.
            * 'BDF': Implicit multi-step variable-order (1 to 5) method based
              on a backward differentiation formula for the derivative
              approximation [5]_. The implementation follows the one described
              in [6]_. A quasi-constant step scheme is used and accuracy is
              enhanced using the NDF modification. Can be applied in the
              complex domain.

        You can also pass an arbitrary class derived from `DaeSolver` which
        implements the solver.
    t_eval : array_like or None, optional
        Times at which to store the computed solution, must be sorted and lie
        within `t_span`. If None (default), use points selected by the solver.
    dense_output : bool, optional
        Whether to compute a continuous solution. Default is False.
    events : callable, or list of callables, optional
        Events to track. If None (default), no events will be tracked.
        Each event occurs at the zeros of a continuous function of time and
        state. Each function must have the signature ``event(t, y)`` where
        additional argument have to be passed if ``args`` is used (see
        documentation of ``args`` argument). Each function must return a
        float. The solver will find an accurate value of `t` at which
        ``event(t, y(t)) = 0`` using a root-finding algorithm. By default,
        all zeros will be found. The solver looks for a sign change over
        each step, so if multiple zero crossings occur within one step,
        events may be missed. Additionally each `event` function might
        have the following attributes:

            terminal: bool or int, optional
                When boolean, whether to terminate integration if this event occurs.
                When integral, termination occurs after the specified the number of
                occurences of this event.
                Implicitly False if not assigned.
            direction: float, optional
                Direction of a zero crossing. If `direction` is positive,
                `event` will only trigger when going from negative to positive,
                and vice versa if `direction` is negative. If 0, then either
                direction will trigger event. Implicitly 0 if not assigned.

        You can assign attributes like ``event.terminal = True`` to any
        function in Python.
    # vectorized : bool, optional
    #     Whether `fun` can be called in a vectorized fashion. Default is False.

    #     If ``vectorized`` is False, `fun` will always be called with ``y`` of
    #     shape ``(n,)``, where ``n = len(y0)``.

    #     If ``vectorized`` is True, `fun` may be called with ``y`` of shape
    #     ``(n, k)``, where ``k`` is an integer. In this case, `fun` must behave
    #     such that ``fun(t, y)[:, i] == fun(t, y[:, i])`` (i.e. each column of
    #     the returned array is the time derivative of the state corresponding
    #     with a column of ``y``).

    #     Setting ``vectorized=True`` allows for faster finite difference
    #     approximation of the Jacobian by methods 'Radau' and 'BDF', but
    #     will result in slower execution for other methods and for 'Radau' and
    #     'BDF' in some circumstances (e.g. small ``len(y0)``).
    args : tuple, optional
        Additional arguments to pass to the user-defined functions.  If given,
        the additional arguments are passed to all user-defined functions.
        So if, for example, `fun` has the signature ``fun(t, y, a, b, c)``,
        then `jac` (if given) and any event functions must have the same
        signature, and `args` must be a tuple of length 3.
    **options
        Options passed to a chosen solver. All options available for already
        implemented solvers are listed below.
    first_step : float or None, optional
        Initial step size. Default is `None` which means that the algorithm
        should choose.
    max_step : float, optional
        Maximum allowed step size. Default is np.inf, i.e., the step size is not
        bounded and determined solely by the solver.
    rtol, atol : float or array_like, optional
        Relative and absolute tolerances. The solver keeps the local error
        estimates less than ``atol + rtol * abs(y)``. Here `rtol` controls a
        relative accuracy (number of correct digits), while `atol` controls
        absolute accuracy (number of correct decimal places). To achieve the
        desired `rtol`, set `atol` to be smaller than the smallest value that
        can be expected from ``rtol * abs(y)`` so that `rtol` dominates the
        allowable error. If `atol` is larger than ``rtol * abs(y)`` the
        number of correct digits is not guaranteed. Conversely, to achieve the
        desired `atol` set `rtol` such that ``rtol * abs(y)`` is always smaller
        than `atol`. If components of y have different scales, it might be
        beneficial to set different `atol` values for different components by
        passing array_like with shape (n,) for `atol`. Default values are
        1e-3 for `rtol` and 1e-6 for `atol`.
    jac : (array_like, array_like), (sparse_matrix, sparse_matrix), callable or None, optional
        Jacobian matrices of the right-hand side of the system with respect
        to y and y', required by the 'Radau' and 'BDF' method. The
        Jacobian matrices have shape (n, n) and their elements (i, j) are equal to
        ``d f_i / d y_j`` and ``d f_i / d y_j'``, respectively.  There are 
        three ways to define the Jacobian:

            * If (array_like, array_like) or (sparse_matrix, sparse_matrix) 
              the Jacobian matrices are assumed to be constant.
            # TODO: Add constant J_y'!
            * If callable, the Jacobians are assumed to depend on both
              t, y and y'; it will be called as ``jac(t, y, y')``, as necessary.
              Additional arguments have to be passed if ``args`` is
              used (see documentation of ``args`` argument).
              For 'Radau' and 'BDF' methods, the return value might be a
              tuple of sparse matrices.
            * If None (default), the Jacobians will be approximated by
              finite differences.

        It is generally recommended to provide the Jacobians rather than
        relying on a finite-difference approximation.
    jac_sparsity : array_like, sparse matrix or None, optional
        Defines a sparsity structure of the Jacobian matrix for a finite-
        difference approximation. Its shape must be (n, n). This argument
        is ignored if `jac` is not `None`. If the Jacobian has only few
        non-zero elements in *each* row, providing the sparsity structure
        will greatly speed up the computations [10]_. A zero entry means that
        a corresponding element in the Jacobian is always zero. If None
        (default), the Jacobian is assumed to be dense.

    Returns
    -------
    Bunch object with the following fields defined:
    t : ndarray, shape (n_points,)
        Time points.
    y : ndarray, shape (n, n_points)
        Values of the solution at `t`.
    sol : `OdeSolution` or None
        Found solution as `OdeSolution` instance; None if `dense_output` was
        set to False.
    t_events : list of ndarray or None
        Contains for each event type a list of arrays at which an event of
        that type event was detected. None if `events` was None.
    y_events : list of ndarray or None
        For each value of `t_events`, the corresponding value of the solution.
        None if `events` was None.
    nfev : int
        Number of evaluations of the right-hand side.
    njev : int
        Number of evaluations of the Jacobian.
    nlu : int
        Number of LU decompositions.
    # TODO: Add number of solve LGS's as done by matlab.
    status : int
        Reason for algorithm termination:

            * -1: Integration step failed.
            *  0: The solver successfully reached the end of `tspan`.
            *  1: A termination event occurred.

    message : string
        Human-readable description of the termination reason.
    success : bool
        True if the solver reached the interval end or a termination event
        occurred (``status >= 0``).

    References
    ----------
    .. [4] E. Hairer, G. Wanner, "Solving Ordinary Differential Equations II:
           Stiff and Differential-Algebraic Problems", Sec. IV.8.
    .. [5] `Backward Differentiation Formula
            <https://en.wikipedia.org/wiki/Backward_differentiation_formula>`_
            on Wikipedia.
    .. [6] L. F. Shampine, M. W. Reichelt, "THE MATLAB ODE SUITE", SIAM J. SCI.
           COMPUTE., Vol. 18, No. 1, pp. 1-22, January 1997.
    .. [7] A. C. Hindmarsh, "ODEPACK, A Systematized Collection of ODE
           Solvers," IMACS Transactions on Scientific Computation, Vol 1.,
           pp. 55-64, 1983.
    .. [8] L. Petzold, "Automatic selection of methods for solving stiff and
           nonstiff systems of ordinary differential equations", SIAM Journal
           on Scientific and Statistical Computing, Vol. 4, No. 1, pp. 136-148,
           1983.
    .. [9] `Stiff equation <https://en.wikipedia.org/wiki/Stiff_equation>`_ on
           Wikipedia.
    .. [10] A. Curtis, M. J. D. Powell, and J. Reid, "On the estimation of
            sparse Jacobian matrices", Journal of the Institute of Mathematics
            and its Applications, 13, pp. 117-120, 1974.
    .. [11] `Cauchy-Riemann equations
             <https://en.wikipedia.org/wiki/Cauchy-Riemann_equations>`_ on
             Wikipedia.
    .. [13] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential
            Equations I: Nonstiff Problems", Sec. II.
    .. [14] `Page with original Fortran code of Radau
            <http://www.unige.ch/~hairer/software.html>`_.


    """
    if method not in METHODS and not (
            inspect.isclass(method) and issubclass(method, DaeSolver)):
        raise ValueError(f"`method` must be one of {METHODS} or DaeSolver class.")

    t0, t_bound = map(float, t_span)

    if args is not None:
        # Wrap the user's fun (and jac, if given) in lambdas to hide the
        # additional parameters. Pass in the original fun as a keyword
        # argument to keep it in the scope of the lambda.
        try:
            _ = [*(args)]
        except TypeError as exp:
            suggestion_tuple = (
                "Supplied 'args' cannot be unpacked. Please supply `args`"
                f" as a tuple (e.g. `args=({args},)`)"
            )
            raise TypeError(suggestion_tuple) from exp

        def fun(t, x, x_dot, fun=fun):
            return fun(t, x, x_dot, *args)
        
        # TODO: Add validate jac here already, since Jacobians are required for all methods!
        jac = options.get('jac')
        if callable(jac):
            options['jac'] = lambda t, x, x_dot: jac(t, x, x_dot, *args)

    if t_eval is not None:
        t_eval = np.asarray(t_eval)
        if t_eval.ndim != 1:
            raise ValueError("`t_eval` must be 1-dimensional.")

        if np.any(t_eval < min(t0, t_bound)) or np.any(t_eval > max(t0, t_bound)):
            raise ValueError("Values in `t_eval` are not within `t_span`.")

        d = np.diff(t_eval)
        if t_bound > t0 and np.any(d <= 0) or t_bound < t0 and np.any(d >= 0):
            raise ValueError("Values in `t_eval` are not properly sorted.")

        if t_bound > t0:
            t_eval_i = 0
        else:
            # Make order of t_eval decreasing to use np.searchsorted.
            t_eval = t_eval[::-1]
            # This will be an upper bound for slices.
            t_eval_i = t_eval.shape[0]

    if method in METHODS:
        method = METHODS[method]

    solver = method(fun, t0, y0, y_dot0, t_bound, vectorized=vectorized, **options)

    if t_eval is None:
        ts = [t0]
        ys = [y0]
    elif t_eval is not None and dense_output:
        ts = []
        ti = [t0]
        ys = []
    else:
        ts = []
        ys = []

    interpolants = []

    if events is not None:
        events, max_events, event_dir = prepare_events(events)
        event_count = np.zeros(len(events))
        if args is not None:
            # Wrap user functions in lambdas to hide the additional parameters.
            # The original event function is passed as a keyword argument to the
            # lambda to keep the original function in scope (i.e., avoid the
            # late binding closure "gotcha").
            events = [lambda t, x, event=event: event(t, x, *args)
                      for event in events]
        g = [event(t0, y0) for event in events]
        t_events = [[] for _ in range(len(events))]
        y_events = [[] for _ in range(len(events))]
    else:
        t_events = None
        y_events = None

    status = None
    while status is None:
        message = solver.step()

        if solver.status == 'finished':
            status = 0
        elif solver.status == 'failed':
            status = -1
            break

        t_old = solver.t_old
        t = solver.t
        y = solver.y

        if dense_output:
            sol = solver.dense_output()
            interpolants.append(sol)
        else:
            sol = None

        if events is not None:
            g_new = [event(t, y) for event in events]
            active_events = find_active_events(g, g_new, event_dir)
            if active_events.size > 0:
                if sol is None:
                    sol = solver.dense_output()

                event_count[active_events] += 1
                root_indices, roots, terminate = handle_events(
                    sol, events, active_events, event_count, max_events,
                    t_old, t)

                for e, te in zip(root_indices, roots):
                    t_events[e].append(te)
                    y_events[e].append(sol(te))

                if terminate:
                    status = 1
                    t = roots[-1]
                    y = sol(t)

            g = g_new

        if t_eval is None:
            ts.append(t)
            ys.append(y)
        else:
            # The value in t_eval equal to t will be included.
            if solver.direction > 0:
                t_eval_i_new = np.searchsorted(t_eval, t, side='right')
                t_eval_step = t_eval[t_eval_i:t_eval_i_new]
            else:
                t_eval_i_new = np.searchsorted(t_eval, t, side='left')
                # It has to be done with two slice operations, because
                # you can't slice to 0th element inclusive using backward
                # slicing.
                t_eval_step = t_eval[t_eval_i_new:t_eval_i][::-1]

            if t_eval_step.size > 0:
                if sol is None:
                    sol = solver.dense_output()
                ts.append(t_eval_step)
                ys.append(sol(t_eval_step))
                t_eval_i = t_eval_i_new

        if t_eval is not None and dense_output:
            ti.append(t)

    message = MESSAGES.get(status, message)

    if t_events is not None:
        t_events = [np.asarray(te) for te in t_events]
        y_events = [np.asarray(ye) for ye in y_events]

    if t_eval is None:
        ts = np.array(ts)
        ys = np.vstack(ys).T
    elif ts:
        ts = np.hstack(ts)
        ys = np.hstack(ys)

    if dense_output:
        if t_eval is None:
            sol = OdeSolution(
                ts, interpolants, #alt_segment=False
                # ts, interpolants, alt_segment=True if method in [BDFDAE] else False
            )
        else:
            sol = OdeSolution(
                ti, interpolants, #alt_segment=False
                # ti, interpolants, alt_segment=True if method in [BDFDAE] else False
            )
    else:
        sol = None

    return OdeResult(t=ts, y=ys, sol=sol, t_events=t_events, y_events=y_events,
                     nfev=solver.nfev, njev=solver.njev, nlu=solver.nlu,
                     status=status, message=message, success=status >= 0)



import numpy as np
from scipy.linalg import qr, solve_triangular


# TODO: Compare this with
# - ddassl.f by Petzold
# - epsode.f by Bryne and Hindmarsh
def select_initial_step(t0, y0, yp0, t_bound, rtol, atol, max_step):
    """Empirically select a good initial step.

    The algorithm is described in [1]_.

    Parameters
    ----------
    t0 : float
        Initial value of the independent variable.
    y0 : ndarray, shape (n,)
        Initial value of the dependent variable.
    yp0 : ndarray, shape (n,)
        Initial value of the dependent variable's derivative.
    t_bound : float
        Final value of the independent variable.
    rtol : float
        Desired relative tolerance.
    atol : float
        Desired absolute tolerance.
    atol : float
        Desired absolute tolerance.
    max_step : float
        Maximum allowed step size.

    Returns
    -------
    h_abs : float
        Absolute value of the suggested initial step.

    References
    ----------
    .. [1] TODO: Find a reference.
    """
    min_step = 0.0
    threshold = atol / rtol
    hspan = abs(t_bound - t0)

    # compute an initial step size h using yp = y'(t0)
    wt = np.maximum(np.abs(y0), threshold)
    rh = 1.25 * np.linalg.norm(yp0 / wt, np.inf) / np.sqrt(rtol)
    h_abs = min(max_step, hspan)
    if h_abs * rh > 1:
        h_abs = 1 / rh
    h_abs = max(h_abs, min_step)
    return h_abs


# TODO:
# - Document this function and reference Shampine2002 accordingly
# - write tests for different problems (including fixed_y0 and fixed_yp0)
# - use sparse LU-decomposition if jac returns sparse matrices
# - use sparse QR-decomposition if available by scipy
def consistent_initial_conditions(fun, jac, t0, y0, yp0, fixed_y0=None, 
                                  fixed_yp0=None, rtol=1e-3, atol=1e-6, 
                                  *args):
    n = len(y0)
    
    if fixed_y0 is None:
        free_y = np.arange(n)
    else:
        free_y = np.where(fixed_y0 == 0)[0]
    
    if fixed_yp0 is None:
        free_yp = np.arange(n)
    else:
        free_yp = np.where(fixed_yp0 == 0)[0]
    
    if len(free_y) + len(free_yp) < n:
        raise ValueError(f"Too many components fixed, cannot solve the problem.")

    if not (isinstance(rtol, float) and rtol > 0):
        raise ValueError("Relative tolerance must be a positive scalar.")
    
    if rtol < 100 * np.finfo(float).eps:
        rtol = 100 * np.finfo(float).eps
        print(f"Relative tolerance increased to {rtol}")
    
    if np.any(np.array(atol) <= 0):
        raise ValueError("Absolute tolerance must be positive.")
    
    y0 = np.asarray(y0).flatten()
    yp0 = np.asarray(yp0).flatten()
    f = fun(t0, y0, yp0, *args)
    Jy, Jyp = jac(t0, y0, yp0, f)
    
    normf0 = np.linalg.norm(f)
    for _ in range(10):
        for _ in range(3):
            dy, dyp = solve_underdetermined_system(f, Jy, Jyp, n, free_y, free_yp)
            
            nrmv = max(np.linalg.norm(np.concatenate([y0, yp0])), np.linalg.norm(atol))
            nrmdv = np.linalg.norm(np.concatenate([dy, dyp]))
            
            if nrmdv > 2 * nrmv:
                factor = 2 * nrmv / nrmdv
                dy *= factor
                dyp *= factor
                nrmdv *= factor
            
            y0 += dy
            yp0 += dyp
            f = fun(t0, y0, yp0, *args)
            fnorm = np.linalg.norm(f)
            
            if (fnorm <= normf0) and (nrmdv <= 1e-3 * rtol * nrmv):
                return y0, yp0, fnorm
        
        Jy, Jyp = jac(t0, y0, yp0, f)
    
    raise RuntimeError("Convergence failed.")


def solve_underdetermined_system(f, Jy, Jyp, n, free_y, free_yp):
    """Solve the underdetermined system 
        0 = f + Jyp @ Delta_yp + Jy @ Delta_y
    A solution is obtained with as many components as possible of 
    (transformed) Delta_yp and Delta_y set to zero.
    """
    Delta_y = np.zeros(n)
    Delta_yp = np.zeros(n)
    
    fixed = (n - len(free_y)) + (n - len(free_yp))
    
    if len(free_y) == 0:
        rank, Q, R, p = qrank(Jyp)
        rankdef = n - rank
        if rankdef > 0:
            if rankdef <= fixed:
                raise ValueError(f"Too many fixed components, rank deficiency is {rankdef}.")
            else:
                raise ValueError("Index greater than one.")
        d = -Q.T @ f
        Delta_yp[free_yp] = np.linalg.solve(R, d)
        return Delta_y, Delta_yp
    
    if len(free_yp) == 0:
        rank, Q, R, p = qrank(Jy)
        rankdef = n - rank
        if rankdef > 0:
            if rankdef <= fixed:
                raise ValueError(f"Too many fixed components, rank deficiency is {rankdef}.")
            else:
                raise ValueError("Index greater than one.")
        d = -Q.T @ f
        Delta_y[free_y] = np.linalg.solve(R, d)
        return Delta_y, Delta_yp
    
    Jy = Jy[:, free_y]
    Jyp = Jyp[:, free_yp]
    
    # QR-decomposition of Fyp leads to the system
    # [R11, R12] [w1'] + [S11, S12] [w1] = [d1]
    # [  0,   0] [w2'] + [S21, S22] [w2] = [d2]
    # with S = Q.T @ Fy
    rank, Q, R, p = qrank(Jyp)
    d = -Q.T @ f
    if rank == n:
        # Full rank (ODE) case: 
        # Set all free dy to zero and solve triangular system below
        Delta_y[free_y] = 0
        Delta_yp[free_yp] = np.linalg.solve(R, d)[p]
    else:
        # Rank deficient (DAE) case:
        S = Q.T @ Jy
        rankS, QS, RS, pS = qrank(S[rank:])
        rankdef = n - (rank + rankS)
        if rankdef > 0:
            if rankdef <= fixed:
                raise ValueError(f"Too many fixed components, rank deficiency is {rankdef}.")
            else:
                raise ValueError("Index greater than one.")

        # compute basic solution of underdetermined system
        # [S21, S22] [w1] = d2
        #            [w2]
        # using column pivoting QR-decomposition
        d2 = d[rank:]
        w_ = np.zeros(n)
        w_[:rankS] = solve_triangular(RS[:rankS, :rankS], (QS.T @ d2[:rankS]))
        w = np.zeros(n)
        w[pS] = w_

        # set w2' = 0 and solve the remaining system
        # [R11] w1' = d1 - [S11, S12] [w1]
        #                             [w2]
        w1p = solve_triangular(R[:rank, :rank], d[:rank] - S[:rank] @ w)
        wp_ = np.concatenate([w1p, np.zeros(len(free_yp) - rank)])
        wp = np.zeros(n)
        wp[p] = wp_

        # store w and wp
        Delta_y[free_y] = w
        Delta_yp[free_yp] = wp
    
    return Delta_y, Delta_yp


def qrank(A):
    Q, R, E = qr(A, pivoting=True)
    tol = max(A.shape) * np.finfo(float).eps * abs(R[0, 0])
    rank = np.sum(abs(np.diag(R)) > tol)
    return rank, Q, R, E



import time
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import solve_ivp
from scipy.optimize._numdiff import approx_derivative
from scipy_dae.integrate import solve_dae, consistent_initial_conditions


"""RHS of stiff van der Pol equation, see mathworks.
References:
-----------
mathworks: https://de.mathworks.com/help/matlab/math/solve-stiff-odes.html
"""

mu = 1e3

def rhs(t, y):
    y1, y2 = y

    yp = np.zeros(2, dtype=y.dtype)
    yp[0] = y2
    yp[1] = mu * (1 - y1 * y1) * y2 - y1

    return yp

def F(t, y, yp):
    return yp - rhs(t, y)


def f(t, z):
    y, yp = z[:2], z[2:]
    return np.concatenate((yp, F(t, y, yp)))

def jac(t, y, yp, f):
    n = len(y)
    z = np.concatenate((y, yp))

    def fun_composite(t, z):
        y, yp = z[:n], z[n:]
        return F(t, y, yp)
    
    J = approx_derivative(lambda z: fun_composite(t, z), 
                            z, method="2-point", f0=f)
    J = J.reshape((n, 2 * n))
    Jy, Jyp = J[:, :n], J[:, n:]
    return Jy, Jyp


if __name__ == "__main__":
    # time span
    t0 = 0
    t1 = 3e3
    t_span = (t0, t1)

    # method = "BDF"
    method = "Radau"

    # initial conditions
    y0 = np.array([2, 0], dtype=float)
    yp0 = rhs(t0, y0)
    z0 = np.concatenate((y0, yp0))

    yp0 = np.zeros_like(y0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    y0, yp0, fnorm = consistent_initial_conditions(F, jac, t0, y0, yp0)
    print(f"y0: {y0}")
    print(f"yp0: {yp0}")
    print(f"fnorm: {fnorm}")

    # solver options
    atol = rtol = 1e-4

    ####################
    # reference solution
    ####################
    start = time.time()
    sol = solve_ivp(rhs, t_span, y0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t_scipy = sol.t
    y_scipy = sol.y
    t = sol.t
    y = sol.y
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    ##############
    # dae solution
    ##############
    start = time.time()
    sol = solve_dae(F, t_span, y0, yp0, atol=atol, rtol=rtol, method=method)
    end = time.time()
    print(f"elapsed time: {end - start}")
    t = sol.t
    y = sol.y
    success = sol.success
    status = sol.status
    message = sol.message
    print(f"success: {success}")
    print(f"status: {status}")
    print(f"message: {message}")
    print(f"nfev: {sol.nfev}")
    print(f"njev: {sol.njev}")
    print(f"nlu: {sol.nlu}")

    # visualization
    fig, ax = plt.subplots(2, 1)

    ax[0].plot(t, y[0], "-ok", label=f"y ({method})", mfc="none")
    ax[0].plot(t_scipy, y_scipy[0], "-xr", label="y scipy")
    ax[0].legend()
    ax[0].grid()

    ax[1].plot(t, y[1], "-ok", label=f"y_dot ({method})", mfc="none")
    ax[1].plot(t_scipy, y_scipy[1], "-xr", label="y_dot scipy")
    ax[1].legend()
    ax[1].grid()

    plt.show()



import numpy as np
from scipy.sparse import issparse, csc_matrix
from scipy.sparse.sputils import isshape
from scipy.optimize._numdiff import group_columns
from scipy.integrate._ivp.common import (
    validate_max_step, validate_tol, num_jac, 
    validate_first_step,
)
from scipy.integrate._ivp.base import ConstantDenseOutput
from scipy.linalg import lu_factor, lu_solve
from scipy.sparse import csc_matrix, issparse, eye
from scipy.sparse.linalg import splu
from .common import select_initial_step


def check_arguments(fun, y0, yp0, support_complex):
    """Helper function for checking arguments common to all solvers."""
    y0 = np.asarray(y0)
    yp0 = np.asarray(yp0)
    if np.issubdtype(np.common_type(y0, yp0), np.complexfloating):
        if not support_complex:
            raise ValueError("`y0` or `yp0` is complex, but the chosen "
                             "solver does not support integration in a "
                             "complex domain.")
        dtype = complex
    else:
        dtype = float
    y0 = y0.astype(dtype, copy=False)
    yp0 = yp0.astype(dtype, copy=False)

    if y0.ndim != 1:
        raise ValueError("`y0` must be 1-dimensional.")
    if yp0.ndim != 1:
        raise ValueError("`yp0` must be 1-dimensional.")
    
    if y0.shape != yp0.shape:
        raise ValueError("`y0` and `yp0` must be of same shape.")

    if not np.isfinite(y0).all():
        raise ValueError("All components of the initial state `y0` must be finite.")
    if not np.isfinite(yp0).all():
        raise ValueError("All components of the initial state `yp0` must be finite.")

    def fun_wrapped(t, y, yp):
        return np.asarray(fun(t, y, yp), dtype=dtype)

    return fun_wrapped, y0, yp0

# TODO:
# - check documentation
# - add consistent initial conditions somehow
class DaeSolver:
    """Base class for DAE solvers.

    In order to implement a new solver you need to follow the guidelines:

        1. A constructor must accept parameters presented in the base class
           (listed below) along with any other parameters specific to a solver.
        2. A constructor must accept arbitrary extraneous arguments
           ``**extraneous``, but warn that these arguments are irrelevant
           using `common.warn_extraneous` function. Do not pass these
           arguments to the base class.
        3. A solver must implement a private method `_step_impl(self)` which
           propagates a solver one step further. It must return tuple
           ``(success, message)``, where ``success`` is a boolean indicating
           whether a step was successful, and ``message`` is a string
           containing description of a failure if a step failed or None
           otherwise.
        4. A solver must implement a private method `_dense_output_impl(self)`,
           which returns a `DenseOutput` object covering the last successful
           step.
        5. A solver must have attributes listed below in Attributes section.
           Note that ``t_old`` and ``step_size`` are updated automatically.
        6. Use `fun(self, t, y, yp)` method for the system evaluation, this
           way the number of function evaluations (`nfev`) will be tracked
           automatically.
        7. For convenience, a base class provides `fun_single(self, t, y, yp)`
           and `fun_vectorized(self, t, y, yp)` for evaluating the system in
           non-vectorized and vectorized fashions respectively (regardless of
           how `fun` from the constructor is implemented). These calls don't
           increment `nfev`.
        8. If a solver uses a Jacobian matrix and LU decompositions, it should
           track the number of Jacobian evaluations (`njev`) and the number of
           LU decompositions (`nlu`).
        9. By convention, the function evaluations used to compute a finite
           difference approximation of the Jacobian should not be counted in
           `nfev`, thus use `fun_single(self, t, y, yp)` or
           `fun_vectorized(self, t, y, yp)` when computing a finite 
           difference approximation of the Jacobian.

    Parameters
    ----------
    fun : callable
        Function defining the DAE system: ``f(t, y, yp) = 0``. The calling 
        signature is ``fun(t, y, yp)``, where ``t`` is a scalar and 
        ``y, yp`` are ndarrays with 
        ``len(y) = len(yp) = len(y0) = len(yp0)``. ``fun`` must return 
        an array of the same shape as ``y, yp``. See `vectorized` for more
        information.
    t0 : float
        Initial time.
    y0 : array_like, shape (n,)
        Initial state.
    yp0 : array_like, shape (n,), optional
        Initial derivative. If the derivative is not given by the user, it is 
        esimated using??? => TODO: See Petzold/ Shampine and coworkers how 
        this is done.
    t_bound : float
        Boundary time --- the integration won't continue beyond it. It also
        determines the direction of the integration.
    var_index : array_like, shape (n,), optional
        Differentiation index of the respective row of f(t, y, y') = 0. 
        Depending on this index, the error estimates are scaled by the 
        stepsize h**(index - 1) in order to ensure convergence.
        Default is None, which means all equations are differential equations.
    vectorized : bool
        Whether `fun` can be called in a vectorized fashion. Default is False.

        If ``vectorized`` is False, `fun` will always be called with ``y`` and
        ``yp`` of shape ``(n,)``, where ``n = len(y0) = len(yp0)``.

        If ``vectorized`` is True, `fun` may be called with ``y`` and 
        ``yp`` of shape ``(n, k)``, where ``k`` is an integer. In this 
        case, `fun` must behave such that 
        ``fun(t, y, yp)[:, i] == fun(t, y[:, i], yp[:, i])`` (i.e. each 
        column of the returned array is the defect of the nonlinear equation 
        corresponding with a column of ``y`` and ``yp``.

        Setting ``vectorized=True`` allows for faster finite difference
        approximation of the Jacobian by methods 'Radau' and 'BDF', but
        will result in slower execution for other methods. It can also
        result in slower overall execution for 'Radau' and 'BDF' in some
        circumstances (e.g. small ``len(y0)``).
    support_complex : bool, optional
        Whether integration in a complex domain should be supported.
        Generally determined by a derived solver class capabilities.
        Default is False.

    Attributes
    ----------
    n : int
        Number of equations.
    status : string
        Current status of the solver: 'running', 'finished' or 'failed'.
    t_bound : float
        Boundary time.
    direction : float
        Integration direction: +1 or -1.
    t : float
        Current time.
    y : ndarray
        Current state.
    yp : ndarray
        Current derivative.
    t_old : float
        Previous time. None if no steps were made yet.
    step_size : float
        Size of the last successful step. None if no steps were made yet.
    nfev : int
        Number of the system's rhs evaluations.
    njev : int
        Number of the Jacobian evaluations.
    nlu : int
        Number of LU decompositions.
    """
    TOO_SMALL_STEP = "Required step size is less than spacing between numbers."

    def __init__(self, fun, t0, y0, yp0, t_bound, rtol, atol, 
                 first_step=None, max_step=np.inf, vectorized=False,
                 jac=None, jac_sparsity=None, support_complex=False):
        self.t_old = None
        self.t = t0
        self._fun, self.y, self.yp = check_arguments(fun, y0, yp0, support_complex)
        self.t_bound = t_bound
        self.vectorized = vectorized
        self.n = self.y.size
        assert self.n == self.yp.size, "`y` and `yp` have to be of same size"

        self.max_step = validate_max_step(max_step)
        self.rtol, self.atol = validate_tol(rtol, atol, self.n)
        if first_step is None:
            self.h_abs = select_initial_step(
                self.t, self.y, self.yp, self.t_bound, 
                self.rtol, self.atol, self.max_step)
        else:
            self.h_abs = validate_first_step(first_step, t0, t_bound)

        self.nfev = 0
        self.njev = 0
        self.nlu = 0
        self._nlusove = 0

        if vectorized:
            def fun_single(t, y, yp):
                return self._fun(t, y[:, None], yp[:, None]).ravel()
            fun_vectorized = self._fun
        else:
            fun_single = self._fun

            def fun_vectorized(t, y, yp):
                if isshape(y.shape) or isshape(yp.shape):
                    if isshape(y.shape):
                        m = y.shape[1]
                        yp = np.tile(yp[:, None], m)
                    else:
                        m = yp.shape[1]
                        y = np.tile(y[:, None], m)
                else:
                    raise RuntimeError("This case should be impossible")
                    m = 1
                    y = np.tile(y[:, None], m)
                    yp = np.tile(yp[:, None], m)

                f = np.empty_like(y)
                for i, (yi, ypi) in enumerate(zip(y.T, yp.T)):
                    f[:, i] = self._fun(t, yi, ypi)
                return f

        def fun(t, y, yp):
            self.nfev += 1
            return self.fun_single(t, y, yp)
        
        self.fun = fun
        self.fun_single = fun_single
        self.fun_vectorized = fun_vectorized
        self.f = self.fun(self.t, self.y, self.yp)

        self.direction = np.sign(t_bound - t0) if t_bound != t0 else 1
        self.status = 'running'

        self.jac_factor_y = None
        self.jac_factor_yp = None
        self.jac, self.Jy, self.Jyp = self._validate_jac(jac, jac_sparsity)

        if issparse(self.Jy) or issparse(self.Jyp):
            def lu(A):
                self.nlu += 1
                return splu(A)

            def solve_lu(LU, b):
                self._nlusove += 1
                return LU.solve(b)

            self.I = eye(self.n, format='csc')
        else:
            def lu(A):
                self.nlu += 1
                return lu_factor(A, overwrite_a=True)

            def solve_lu(LU, b):
                self._nlusove += 1
                return lu_solve(LU, b, overwrite_b=True)

            self.I = np.identity(self.n)
        self.lu = lu
        self.solve_lu = solve_lu
    
    def _validate_jac(self, jac, sparsity):
        t0 = self.t
        y0 = self.y
        yp0 = self.yp
        dtype = np.common_type(y0, yp0)

        if jac is None:
            if sparsity is not None:
                sparsity_y, sparsity_yp = sparsity
                if issparse(sparsity_y):
                    sparsity_y = csc_matrix(sparsity_y)
                if issparse(sparsity_yp):
                    sparsity_yp = csc_matrix(sparsity_yp)
                groups_y = group_columns(sparsity_y)
                groups_yp = group_columns(sparsity_yp)
                sparsity_y = (sparsity_y, groups_y)
                sparsity_yp = (sparsity_yp, groups_yp)
            else:
                sparsity_y, sparsity_yp = None, None

            def jac_wrapped(t, y, yp, f):
                self.njev += 1
                # Jy, self.jac_factor_y = num_jac(
                #     lambda t, y: self.fun_vectorized(t, y, np.tile(yp[:, None], self.n)), 
                #     t, y, f, self.atol, self.jac_factor_y, sparsity_y)
                # Jyp, self.jac_factor_yp = num_jac(
                #     lambda t, yp: self.fun_vectorized(t, np.tile(y[:, None], self.n), yp), 
                #     t, yp, f, self.atol, self.jac_factor_yp, sparsity_y)
                # n_vec = max(y.shape[1], yp.shape[1])
                # if isshape(y.shape):
                #     ny, my = y.shape
                # else:
                #     ny, my = self.n, 1
                # if isshape(yp.shape):
                #     nyp, myp = yp.shape
                # else:
                #     nyp, myp = self.n, 1
                # n, m = max(ny, nyp), max(my, myp)

                # y = np.tile(y[:, None], )
                # yp = np.atleast_2d(yp)

                # y = np.atleast_2d(y)
                # yp = np.atleast_2d(yp)
                Jy, self.jac_factor_y = num_jac(
                    lambda t, y: self.fun_vectorized(t, y, yp), 
                    t, y, f, self.atol, self.jac_factor_y, sparsity_y)
                Jyp, self.jac_factor_yp = num_jac(
                    lambda t, yp: self.fun_vectorized(t, y, yp), 
                    t, yp, f, self.atol, self.jac_factor_yp, sparsity_y)
                
                return Jy, Jyp
            
            Jy, Jyp = jac_wrapped(t0, y0, yp0, self.f)
        
        elif callable(jac):
            Jy, Jyp = jac(t0, y0, yp0)
            self.njev = 1
            if issparse(Jy) or issparse(Jyp):
                Jy = csc_matrix(Jy, dtype=dtype)
                Jyp = csc_matrix(Jyp, dtype=dtype)

                def jac_wrapped(t, y, yp, _=None):
                    self.njev += 1
                    Jy, Jyp = jac(t, y, yp)
                    Jy = csc_matrix(Jy, dtype=dtype)
                    Jyp = csc_matrix(Jyp, dtype=dtype)
                    return Jy, Jyp

            else:
                Jy = np.asarray(Jy, dtype=dtype)
                Jyp = np.asarray(Jyp, dtype=dtype)

                def jac_wrapped(t, y, yp, _=None):
                    self.njev += 1
                    Jy, Jyp = jac(t, y, yp)
                    Jy = np.asarray(Jy, dtype=dtype)
                    Jyp = np.asarray(Jyp, dtype=dtype)
                    return Jy, Jyp

            if Jy.shape != (self.n, self.n):
                raise ValueError("`Jy` is expected to have shape {}, but "
                                 "actually has {}."
                                 .format((self.n, self.n), Jy.shape))
            if Jyp.shape != (self.n, self.n):
                raise ValueError("`Jyp` is expected to have shape {}, but "
                                 "actually has {}."
                                 .format((self.n, self.n), Jyp.shape))
        
        else:
            Jy, Jyp = jac
            if issparse(Jy) or issparse(Jyp):
                Jy = csc_matrix(Jy, dtype=dtype)
                Jyp = csc_matrix(Jyp, dtype=dtype)
            else:
                Jy = np.asarray(Jy, dtype=dtype)
                Jyp = np.asarray(Jyp, dtype=dtype)

            if Jy.shape != (self.n, self.n):
                raise ValueError("`Jy` is expected to have shape {}, but "
                                 "actually has {}."
                                 .format((self.n, self.n), Jy.shape))
            if Jyp.shape != (self.n, self.n):
                raise ValueError("`Jyp` is expected to have shape {}, but "
                                 "actually has {}."
                                 .format((self.n, self.n), Jyp.shape))
            jac_wrapped = None

        return jac_wrapped, Jy, Jyp

    @property
    def step_size(self):
        if self.t_old is None:
            return None
        else:
            return np.abs(self.t - self.t_old)

    def step(self):
        """Perform one integration step.

        Returns
        -------
        message : string or None
            Report from the solver. Typically a reason for a failure if
            `self.status` is 'failed' after the step was taken or None
            otherwise.
        """
        if self.status != 'running':
            raise RuntimeError("Attempt to step on a failed or finished "
                               "solver.")

        if self.n == 0 or self.t == self.t_bound:
            # Handle corner cases of empty solver or no integration.
            self.t_old = self.t
            self.t = self.t_bound
            message = None
            self.status = 'finished'
        else:
            t = self.t
            success, message = self._step_impl()

            if not success:
                self.status = 'failed'
            else:
                self.t_old = t
                if self.direction * (self.t - self.t_bound) >= 0:
                    self.status = 'finished'

        return message

    def dense_output(self):
        """Compute a local interpolant over the last successful step.

        Returns
        -------
        sol : `DenseOutput`
            Local interpolant over the last successful step.
        """
        if self.t_old is None:
            raise RuntimeError("Dense output is available after a successful "
                               "step was made.")

        if self.n == 0 or self.t == self.t_old:
            # Handle corner cases of empty solver and no integration.
            return ConstantDenseOutput(self.t_old, self.t, self.y)
        else:
            return self._dense_output_impl()

    def _step_impl(self):
        raise NotImplementedError

    def _dense_output_impl(self):
        raise NotImplementedError
